﻿-- Author      : RisM
-- Create Date : 6/28/2009 3:56:24 PM

local SpeakinSpell = LibStub("AceAddon-3.0"):GetAddon("SpeakinSpell")
local L	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell", false)
local DEFAULT_SPEECHES = LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell_DEFAULT_SPEECHES",	false)

SpeakinSpell:PrintLoading("data.lua")

-------------------------------------------------------------------------------
-- DATA STRUCTURES (? not sure what to call this category)
-------------------------------------------------------------------------------


-------------------------------------------------------------------------------
-- INITIALIZATION / IMPORT OF DEFAULT DATA
-------------------------------------------------------------------------------


function SpeakinSpell:InitDefaultSavedData()
	
	SpeakinSpellSavedData.Version = SpeakinSpell.CURRENT_VERSION
	
	self:ValidateObject( SpeakinSpellSavedData, SpeakinSpell.DEFAULTS.GLOBALOPTIONS )

	-- reset/create the Spell Id Cache used for making <spelllink>
	SpeakinSpellSavedDataForAll.SpellIdCache = nil
	SpeakinSpellSavedDataForAll.SpellIdCache = {}

	-- create/reset the EventTable
	SpeakinSpellSavedData.EventTable = nil
	SpeakinSpellSavedData.EventTable = {}

	-- copy defaults into current settings using class-specific and race-specific defaults
	local playerClass, englishClass = UnitClass("player"); --englishClass is the UPPERCASE class name in English
	local playerRace, englishRace = UnitRace("player"); 
	englishRace = string.upper(englishRace) -- set UPPERCASE for table keys below
	
	-- Insert defaults speeches and event settings for: All, Race, and Class defaults
	self:AddTable( SpeakinSpellSavedData.EventTable, DEFAULT_SPEECHES.ALL )
	self:AddTable( SpeakinSpellSavedData.EventTable, DEFAULT_SPEECHES.RACES[englishRace] )
	self:AddTable( SpeakinSpellSavedData.EventTable, DEFAULT_SPEECHES.CLASSES[englishClass] )
	self:AddTable( SpeakinSpellSavedData.EventTable, DEFAULT_SPEECHES.RACECLASSCOMBOS[englishRace.." "..englishClass] )
	
	-- remove faction-specific spells from the defaults (i.e. mage portals) 
	for key, EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		if EventTableEntry.Faction ~= nil then
			if EventTableEntry.Faction ~= UnitFactionGroup("player") then
				SpeakinSpellSavedData.EventTable[key] = nil
			end
		end
	end
	
	-- Reset the NewEventsDetected table --TODO: move this to SpeakinSpellSavedDataForAll
	SpeakinSpellSavedData.NewEventsDetected = nil
	SpeakinSpellSavedData.NewEventsDetected = {}

	-- create the random substitution tables
	SpeakinSpellSavedData.RandomSubs = nil
	SpeakinSpellSavedData.RandomSubs = {}
	self:AddTable( SpeakinSpellSavedData.RandomSubs, DEFAULT_SPEECHES.RandomSubs )
	self:AddRandomFactions() -- Add guild and arena teams to the random substitution tables
	-- TODO: Add "Enemy Of" <guild> and arena teams to <randomtaunt> list when loading defaults
	
	-- some info may be missing from the defaults out of convenience
	self:ValidateAllSavedData()
end


function SpeakinSpell:ImportNewSpeechesFor(key, Messages)
	local funcname = "ImportNewSpeechesFor"
	-- add any new messages from the defaults into our current settings
	-- don't import duplicates
	-- and don't overwrite by numeric index keys like CopyTable or AddTable would do

	-- we will copy from Messages into Target
	local Target = SpeakinSpellSavedData.EventTable[key].Messages

	local ShowedName = false

	for _,newmsg in pairs(Messages) do
		-- see if we already use this message
		local found = false
		for _,oldmsg in pairs(Target) do
			if oldmsg == newmsg then
				found = true
			end
		end

		if not found then
			-- this is a new message

			-- use the chat window to report what we're adding
			-- TODO: add prompts
			if not ShowedName then
				local de = SpeakinSpellSavedData.EventTable[key].DetectedEvent
				self:Print( self:FormatDisplayName( de,nil,true,false ) )
				ShowedName = true
			end

			--self:Print(newmsg)
			tinsert(Target, newmsg)
		end
	end
end


function SpeakinSpell:ImportNewDefaults()
	local funcname = "ImportNewDefaults"

	local playerClass, englishClass = UnitClass("player"); --englishClass is the UPPERCASE class name in English
	local playerRace, englishRace = UnitRace("player");
	englishRace = string.upper(englishRace) -- for table keys

	self:DebugMsgDumpString("playerRace",playerRace)
	self:DebugMsgDumpString("englishRace",englishRace)

	NEW_DEF_MSG_SOURCES = {
		DEFAULT_SPEECHES.ALL,
		DEFAULT_SPEECHES.RACES[englishRace],
		DEFAULT_SPEECHES.CLASSES[englishClass],
	}

	self:Print(L["Adding new default samples to your saved settings"])

	for _,NewEventTable in pairs(NEW_DEF_MSG_SOURCES) do
		for key, NewEvent in pairs(NewEventTable) do
			self:DebugMsg(funcname, "key:"..key)
			if (NewEvent.Faction == nil) or (NewEvent.Faction == UnitFactionGroup("player")) then 
				-- the new event applies to our faction so use it
				if SpeakinSpellSavedData.EventTable[key] then
					-- we already have settings for this event, so merge them
					-- add any new messages from the defaults into our current settings
					self:ImportNewSpeechesFor(key, NewEvent.Messages)
				else
					-- we don't have settings for this event
					-- add a copy of this new default event to the current settings
					-- TODO: add prompts
					SpeakinSpellSavedData.EventTable[key] = self:CopyTable(NewEventTable[key])
					
					-- validate and display the DisplayName
					local de = SpeakinSpellSavedData.EventTable[key].DetectedEvent
					self:Print( self:FormatDisplayName( de,nil,true,false ) )
					--self:Print(L["(new event - all speeches are new)"])
				end
			end
		end
	end

	-- import random substitutions
	-- TODO: add prompts
	self:AddTable( SpeakinSpellSavedData.RandomSubs, DEFAULT_SPEECHES.RandomSubs )
	-- Add guild and arena teams to the <randomfaction> substitution when importing defaults
	-- TODO: add prompts
	self:AddRandomFactions()
	
	--TODO: change InitDefaultSavedData to reset to empty tables and then ImportNewDefaults with prompts disabled / silent mode

	self:Print(L["Import complete"])
end


-------------------------------------------------------------------------------
-- DATA STRUCTURES (? not sure what to call this category)
-------------------------------------------------------------------------------

function SpeakinSpell:InitRuntimeData()
	--NOTE: SpeakinSpellSavedData either does not exist yet or is about to be reset to defaults
	--		do not reference it here
	
	-- this is just a declaration
	self.RuntimeData = {
		AnnouncementHistory = {}, -- [eventkey] = {LastMessage, LastTime, LastTarget}
		AnnouncedThisCombat = {}, -- [eventkey] = true, for Limit Once Per Combat - this table is destroyed/reset on enter/exit combat
		LastRandomSub = {}, -- [randomkey] = "random value" last used for the <randomkey> substitution
		OptionsGUIStates = {
			MessageSettings = {
				SelectedEventKey = nil,
				FilterChanged = true,
			},
			CreateNew = {
				SelectedEventKey = nil,
				FilterChanged = true,
			},
			RandomSubs = {
				SelectedSub = nil,
			},
			LastPageViewed = {
				ObjectName = "General",
				InterfacePanel = L["SpeakinSpell"],
			},
			SelectedEventTypeFilter = "ALL",
			SelectedEventTextFilter = "",
		},
	}
end


function SpeakinSpell:DeleteSpell(key)
	-- delete the current selection from the SavedData and RuntimeData tables
	SpeakinSpellSavedData.EventTable[key] = nil
	
	if not self.RuntimeData then
		-- we're still initializing, so the rest of this stuff doesn't matter
		return
	end
	
	-- remove runtime data for messaging cooldowns and limits
	self.RuntimeData.AnnouncementHistory[key] = nil
	
	-- restore the spell to the list of new detected
	-- so the user doesn't have to retrigger it
	-- self:RecordNewSpell(EventTableEntry.DetectedEvent)
	-- NO! could be outdated and bugged event
	-- just make the user recreate it	
	
	self:CurrentMessagesGUI_ValidateSelectedEvent() -- make sure we still have a valid spell selection
end



function SpeakinSpell:IsEventTypeRankable(checktype)
	local RankableEventTypes = {
		"ALL",
		"UNIT_SPELLCAST_SENT",
		"UNIT_SPELLCAST_CHANNEL_START",
		"UNIT_SPELLCAST_CHANNEL_STOP",
		"UNIT_SPELLCAST_FAILED",
		"UNIT_SPELLCAST_INTERRUPTED",
		"UNIT_SPELLCAST_STOP",
		"UNIT_SPELLCAST_SUCCEEDED",
	}
	for _,type in pairs(RankableEventTypes) do
		if checktype == type then
			return true --this event type is rankable
		end
	end
	
	return false --this event type is not rankable
end


function SpeakinSpell:IsSelectedEventTypeFilterRankable()
	return self:IsEventTypeRankable( self.RuntimeData.OptionsGUIStates.SelectedEventTypeFilter )
end


function SpeakinSpell:SetFilterShowMoreThanAHundred(value)
	if SpeakinSpellSavedData.ShowMoreThanAHundred == value then
		--value is not changing
		return
	end
	
	SpeakinSpellSavedData.ShowMoreThanAHundred = value
	
	--self.RuntimeData.OptionsGUIStates.MessageSettings.FilterChanged = true --the message settings GUI will always Show All Ranks
	self.RuntimeData.OptionsGUIStates.CreateNew.FilterChanged = true
end


function SpeakinSpell:SetFilterShowAllRanks(value)
	if SpeakinSpellSavedData.ShowAllRanks == value then
		--value is not changing
		return
	end
	
	SpeakinSpellSavedData.ShowAllRanks = value
	
	if self:IsSelectedEventTypeFilterRankable() then
		--self.RuntimeData.OptionsGUIStates.MessageSettings.FilterChanged = true --the message settings GUI will always Show All Ranks
		self.RuntimeData.OptionsGUIStates.CreateNew.FilterChanged = true
	end
end


function SpeakinSpell:SetFilter(type,text)
	local rebuild = false
	if type and self.RuntimeData.OptionsGUIStates.SelectedEventTypeFilter ~= type then
		self.RuntimeData.OptionsGUIStates.SelectedEventTypeFilter = type
		rebuild = true
	end
	
	if text and self.RuntimeData.OptionsGUIStates.SelectedEventTextFilter ~= text then
		self.RuntimeData.OptionsGUIStates.SelectedEventTextFilter = text
		rebuild = true
	end
	
	if rebuild then
		self.RuntimeData.OptionsGUIStates.MessageSettings.FilterChanged = true
		self.RuntimeData.OptionsGUIStates.CreateNew.FilterChanged = true
	end
end

	
function SpeakinSpell:MatchesFilter(DetectedEvent, ForceAllRanks)
	local funcname = "MatchesFilter"
		
	-- check type filter
	local type = SpeakinSpell.RuntimeData.OptionsGUIStates.SelectedEventTypeFilter
	if type and not (type == "ALL" or type == DetectedEvent.type) then
		--self:DebugMsg(funcname, "type failed:"..DetectedEvent.type)
		return false
	end
	
	-- check text filter - use lowercase for case-insensitive comparison
	local text = string.lower( SpeakinSpell.RuntimeData.OptionsGUIStates.SelectedEventTextFilter )
	--self:DebugMsg(funcname, "text:"..text)
	if text and not (text == "") then
		local name = string.lower(DetectedEvent.name)
		-- KLUGE: hyphens apparently cause string.find to fail to find "auto-afk" in "auto-afk"
		--		so replace all hyphens with spaces for the sake of comparison
		text = string.gsub(text,"-"," ")
		name = string.gsub(name,"-"," ")
		-- look for the text as a substring anywhere in the name
		local ndxMatch = string.find( name, text )
		if not (ndxMatch and (ndxMatch >= 0)) then
			-- the substring was not found, so this is not a match
			--self:DebugMsg(funcname, string.format("text \"%s\" failed on name \"%s\"",text,name))
			return false
		end
		-- else, don't return true yet because we still need to check the rank filter
	end
	
	-- check rank filter
	if self:IsSelectedEventTypeFilterRankable() then
		local HideSpecialRanks = not (ForceAllRanks or SpeakinSpellSavedData.ShowAllRanks)
		if HideSpecialRanks then
			if DetectedEvent.rank and (DetectedEvent.rank ~= "") then
				-- a rank is specified
				--self:DebugMsg(funcname, "rank failed:"..DetectedEvent.rank)
				return false
			end
		end
	end

	-- no filters are defined
	--self:DebugMsg(funcname, "pass key:"..DetectedEvent.key)
	return true
end


function SpeakinSpell:ColorFilterText( DisplayName, ReturnColor )
	-- do a case insensitive match to color-code
	local filter = string.lower( SpeakinSpell.RuntimeData.OptionsGUIStates.SelectedEventTextFilter )
	if filter and not (filter == "") then
		local name = string.lower( DisplayName )
		local ndx = string.find( name, filter )
		if ndx == nil then
			return DisplayName
		end
		
		if not ReturnColor then
			ReturnColor = "|r"
		end
		
		-- split up the string
		local Start = string.sub( DisplayName, 1, ndx - 1 )
		local Match = string.sub( DisplayName,    ndx, ndx + string.len(filter) - 1 )
		local Rest  = string.sub( DisplayName,         ndx + string.len(filter) )
		
		-- put it back together again
		return Start..(SpeakinSpell.Colors.SearchMatch)..Match..(ReturnColor)..Rest
	end
	-- else no search in effect
	return DisplayName
end


-- RecordNewEvent creates and maintains a list of new spells that don't have messages associated with them
-- this list is used to present the user with a list of options when she wants to start announcing a new spell
function SpeakinSpell:RecordNewEvent(de)
	local funcname = "RecordNewEvent"
	
	-- if we already recorded this spell, don't add a duplicate, unless this is a different rank
	-- NOTE: RecordNewEvent is called for both ranked, and rankless versions of the event
	
	if SpeakinSpellSavedData.NewEventsDetected[de.key] then
		-- this is not a new event, so we're done
		--self:DebugMsg(funcname, self:FormatSubs("New Event seen before <key>: <SpellLink>", de) )
		self:CreateNew_OnNewEventDetected(de)
		return
	end
	
	-- record the new event
	self:DebugMsg(funcname, self:FormatSubs("Found New Event <key>: <SpellLink>", de) )
	SpeakinSpellSavedData.NewEventsDetected[de.key] = de
	
	-- update the options GUI
	self:CreateNew_OnNewEventDetected(de)
end


-- The only way to find a spell ID for a name is to search all possible IDs for a match
-- because of this, we use a cache table to speed up the search for known spells
function SpeakinSpell:GetSpellID( spellname, spellrank )
	local funcname = "GetSpellID"
	
	-- make sure the rank is non-nil
	if not spellrank then
		spellrank = ""
	end
	
	-- determine the key for our cache table
	local cachekey = spellname..spellrank

	-- check for this name in the cache
	local id = SpeakinSpellSavedDataForAll.SpellIdCache[ cachekey ]
	if id ~= nil then -- NOTE: allow the 0 id for non-spell events
		return id
	end

	-- search all possible spellIds for a match
	local name = ""
	local rank = ""

	for i = 1, 100000 do
		name,rank = GetSpellInfo(i)
		if name then
			local thiskey = name..rank
			if thiskey == cachekey then
				self:DebugMsg(funcname, string.format("new spellid cachekey:%s=%s", cachekey,tostring(i)) )
				SpeakinSpellSavedDataForAll.SpellIdCache[ cachekey ] = i
				return i
			end
		end
	end

	-- no spell id found
	self:DebugMsg(funcname, string.format("new spellid cachekey:%s=%s", cachekey,tostring(0)) )
	SpeakinSpellSavedDataForAll.SpellIdCache[ cachekey ] = 0
	return 0
end



-----------------------------------------------------------------------------------
-- ValidateSpellIdCache -- Validate the saved SpellId Cache
--
--	This should only be necessary if Blizzard changes the spell ids
--		but is also a general data validation task, in case anything has been corrupted
--		or in case I have programming errors
--		this function iterates the spell id cache to make sure that the names still match the IDs
--
--	if Blizzard ever changes the spell ids, then our saved cache would become invalid
--		that's unlikely
--
--	but if our saved data gets messed up from manual edits or some other problems
--		make sure it isn't saving incorrect information
--
function SpeakinSpell:ValidateSpellIdCache()
	-- if the cache does not exist, create it and we're done
	if not SpeakinSpellSavedDataForAll.SpellIdCache then
		SpeakinSpellSavedDataForAll.SpellIdCache = {}
		return
	end
	
	-- check every entry in the table 
	-- should take fewer iterations that way than for i = 1, 100000 do ...
	-- the id should be usable to query the name
	-- which should be usable to determine the key
	-- which should return the same id we started with, or we need to fix it
	
	local name = ""
	local rank = ""
	for cachekey, id in pairs(SpeakinSpellSavedDataForAll.SpellIdCache) do
		name,rank = GetSpellInfo(id)
		if not name then -- this id is no longer valid
			SpeakinSpellSavedDataForAll.SpellIdCache[cachekey] = nil
		else
			local newkey = name..rank
			if newkey ~= cachekey then
				self:DebugMsg(funcname, string.format("cachekey:%s, newkey:%s, id:%d", cachekey, newkey, id))
				SpeakinSpellSavedDataForAll.SpellIdCache[cachekey] = nil
				SpeakinSpellSavedDataForAll.SpellIdCache[newkey] = id
			end
		end
	end
end



-- NOTE: GetSpellLink on a name will throw a LUA error on invalid names
--		which is why we had to switch over to spell ids
function SpeakinSpell:SafeGetSpellLink(de)
	-- NOTE: spellid may be either zero or nil if this event is not a spell with an id
	if not de.spellid then
		return "["..de.name.."]"
	end

	local SpellLink = GetSpellLink( de.spellid )
	if SpellLink then
		return SpellLink
	else
		return "["..de.name.."]"
	end
end



-- returns rankedkey, ranklesskey
-- if not a rankable event, then ranklesskey will be nil
-- TODOFUTURE: this could return an ordered list of potential keys?
-- maybe for linking UNIT_SPELLCAST_SENT to the matching SPELL_AURA_GAIN 
-- and/or for making speech sets that apply to many spells with the same speech settings
function SpeakinSpell:MakeEventTableKeys( de )
	-- create a unique key for this event
	local rankedkey		= string.gsub( string.upper(de.type..de.name..de.rank), " ", "_" ) -- ACTIVE 3.1.2.07 format
	local ranklesskey	= string.gsub( string.upper(de.type..de.name),			" ", "_" ) -- ACTIVE 3.1.2.07 format
	
	if self:IsEventTypeRankable( de.type ) and (rankedkey ~= ranklesskey) then
		return rankedkey, ranklesskey
	else
		return rankedkey, nil
	end
end


function SpeakinSpell:CreateDetectedEvent( DetectedEventStub )
	local funcname = "CreateDetectedEvent"
	
	local de = self:CopyTable(DetectedEventStub)

	if not de.caster then
		de.caster = UnitName("player")
	end
	
	-- validate the target - get a backup target from selection, focus, or assumed self-cast
	if not de.target or de.target == "" or de.target == "Unknown" then
		de.target = self:GetDefaultTarget(true)
	end
	
	-- fill in any other missing defaults or auto-generated properties
	self:Validate_DetectedEvent(de)
	
	return de
end


-------------------------------------------------------------------------------
-- CHECK OPTIONS IF SPEAKING IS ALLOWED
-------------------------------------------------------------------------------


function SpeakinSpell:SaveAnnouncementHistory( de, msg )
	-- record info about the last time we announced this event (which should be right now)
	self.RuntimeData.AnnouncementHistory[de.key] = {
		LastMessage = msg,
		LastTime	= GetTime(),
		LastTarget	= de.target,
	}
	self.RuntimeData.AnnouncedThisCombat[de.key] = true
end


function SpeakinSpell:GetLastMessage( de )
	local History = self.RuntimeData.AnnouncementHistory[de.key]
	if History then
		return History.LastMessage
	else
		return nil
	end
end


function SpeakinSpell:CheckSpellFrequency( de )
	-- check the frequency setting against a random number to determine if we should speak or not
	local Chance = math.random(1,100)
	local Frequency = de.EventTableEntry.Frequency*100

	if Chance > Frequency then
		de.chance = Chance
		de.frequency = Frequency
		self:DebugMsg( "CheckSpellFrequency", self:FormatSubs("random chance failed for <SpellLink> rolled <Chance> > <Frequency>", de) )
		return false
	end
	
	return true
end


function SpeakinSpell:CheckSpellCooldown( de )
	local funcname = "CheckSpellCooldown"
	
	-- check for a cooldown limit on announcing this spell

	local History = self.RuntimeData.AnnouncementHistory[de.key]
	if not History then
		-- we have not yet announced this spell, so it can't be on cooldown
		return true
	end

	-- calculate elapsed time since last announcement
	-- NOTE: History.LastTime is guaranteed to be a valid value, or we would have aborted above
	local elapsed = GetTime() - History.LastTime

	-- Now finally actually check if the cooldown is in effect
	-- NOTE: "no cooldown" is set by Cooldown=0, so elapsed < 0 will not occur
	if elapsed < de.EventTableEntry.Cooldown then
		-- the cooldown is in effect, so be silent
		de.elapsed = elapsed
		de.cd = de.EventTableEntry.Cooldown
		self:DebugMsg( funcname, self:FormatSubs("messaging on cooldown <SpellLink> elapsed:<elapsed>, CD:<cd>", de) )
		return false
	end
	
	return true
end


function SpeakinSpell:CheckOncePerTarget( de )
	local funcname = "CheckOncePerTarget"

	if not de.EventTableEntry.OncePerTarget then
		-- "limit once per target" is disabled, so we can always speak
		return true
	end

	local History = self.RuntimeData.AnnouncementHistory[de.key]
	if not History then
		-- we haven't announced this event yet at all, so we can't have announced it for this target
		return true
	end
	
	-- check the last target that was used
	if de.target == History.LastTarget then
		self:DebugMsg( funcname, "Limit once per target is in effect for target="..tostring(de.target) )
		return false
	end
	
	-- we have a new target this time, so we can speak
	return true
end


function SpeakinSpell:CheckOncePerCombat( de )
	local funcname = "CheckOncePerCombat"

	if not de.EventTableEntry.OncePerCombat then
		-- "limit once per combat" is disabled, so we can always speak
		return true
	end
	
	local History = self.RuntimeData.AnnouncementHistory[de.key]
	if not History then
		-- we haven't announced this event yet at all, so we can't have announced it for this target
		return true
	end

--	if not self.RuntimeData.InCombat then
--		-- we're not in combat, so this option should not apply
--		self:DebugMsg( funcname, "Limit once per combat no has no effect because you are not in combat" )
--		return true
--	end
	
	if self.RuntimeData.AnnouncedThisCombat[de.key] then
		self:DebugMsg( funcname, "Limit once per combat is in effect for this event key="..tostring(de.key) )
		return false
	end
	
	-- it's a new combat, so go ahead and announce
	return true
end


function SpeakinSpell:ResetOncePerCombatFlags()
	-- forget all the once-per-combat memory
	-- NOTE: this is easier if we track it as a separate table
	--		rather than add data to self.RuntimeData.AnnouncementHistory
	self.RuntimeData.AnnouncedThisCombat = {}
	
	-- NOTE: don't reset last target info when changing combat sessions
	-- because we want "limit once per target" to apply to targets to the same name in multiple pulls
end


function SpeakinSpell:AllowSpeakForSpell( de )
	local funcname = "AllowSpeakForSpell"
	
	if de.EventTableEntry.DisableAnnouncements then
		self:DebugMsg(funcname, self:FormatSubs("messages for <SpellLink> are disabled", de) )
		return false
	end
	
	if not SpeakinSpellSavedData.EnableAllMessages then
		self:DebugMsg(funcname,"all SpeakinSpell messages are disabled")
		return false
	end
	
	if not self:CheckSpellFrequency( de ) then
		return false
	end
	
	if not self:CheckSpellCooldown( de ) then
		return false
	end
	
	if not self:CheckOncePerTarget( de ) then
		return false
	end
	
	if not self:CheckOncePerCombat( de ) then
		return false
	end
	
	-- no limits or restrictions apply, so go ahead and speak
	return true
end



-- will construct or repair a DetectedEvent (de) object
function SpeakinSpell:Validate_DetectedEvent(de)
	if not SpeakinSpell.DefaultDetectedEvent then
		-- every item in this list must EXIST in a valid DetectedEvent object
		SpeakinSpell.DefaultDetectedEvent = { --TODO: copy this to SpeakinSpellSavedData and add a GUI for it
			name = "Unknown Event",
			rank = "",
			type = "EVENT",
		}
	end
	
	self:ValidateObject( de, SpeakinSpell.DefaultDetectedEvent )

	-- de.type is an enumerated table key that must be a valid one of a limited list
	-- if we don't have locale strings for this type, set to the generic "EVENT" type
	if not L.EVENTTYPES.IN_SPELL_LIST[de.type] then
		de.type = "EVENT"
	end
end


-- will construct or repair an EventTableEntry (ete) object
function SpeakinSpell:Validate_EventTableEntry( ete )
	if not SpeakinSpell.DefaultEventTableEntry then
		-- every item in this list must EXIST in a valid EventTableEntry object
		SpeakinSpell.DefaultEventTableEntry = { --TODO: copy this to SpeakinSpellSavedData and add a GUI for it
			Channels = self:CopyTable(SpeakinSpell.DEFAULTS.CHANNEL_RULES),
			Messages = self:CopyTable(DEFAULT_SPEECHES.NEWSPELL),
			DisableAnnouncements = false,
			WhisperTarget = false,
			Frequency = SpeakinSpell.DEFAULTS.FREQUENCY,
			Cooldown  = SpeakinSpell.DEFAULTS.COOLDOWN,
		}
	end
	
	self:Validate_DetectedEvent( ete.DetectedEvent )
	self:ValidateObject( ete, SpeakinSpell.DefaultEventTableEntry )
end


function SpeakinSpell:ValidateAllSavedData()
	local funcname = "ValidateAllSavedData"
	
	-----------------------------------------------------------------------------------
	-- prior to v3.0.3.08, a version number was not saved in the SpeakinSpellSavedData
	-- create an older version number
	if not SpeakinSpellSavedData.Version then
		SpeakinSpellSavedData.Version = "3.0.3.07"
	end
	
	-----------------------------------------------------------------------------------
	-- Apply patches if necessary
	
	-- if the saved data was from the current version
	-- we can speed up loading a little by skipping this loop
	if SpeakinSpellSavedData.Version < SpeakinSpell.CURRENT_VERSION then
		self:ApplyPatches()
	end
	
	-- global settings - make sure they all exist
	self:ValidateObject( SpeakinSpellSavedData, SpeakinSpell.DEFAULTS.GLOBALOPTIONS )


	-----------------------------------------------------------------------------------
	-- Validate the saved SpellId Cache
	-- This should only be necessary if Blizzard changes the spell ids
	-- but is also a general data validation task, in case anything has been corrupted
	-- or in case I have programming errors
	self:ValidateSpellIdCache()
	
	
	-----------------------------------------------------------------------------------
	-- Apply standard error checks and display string updates
	-- even if running the same version as the saved data
	
	-- remove faction-specific spells (i.e. mage portals) 
	-- and reset runtime data such as localizable display names
	for key, ete in pairs(SpeakinSpellSavedData.EventTable) do
		ete.key = key
		ete.DetectedEvent.key = key
		self:Validate_EventTableEntry(ete)
	end
	
	for key, de in pairs(SpeakinSpellSavedData.NewEventsDetected) do
		de.key = key
		self:Validate_DetectedEvent(de)
	end
end


-------------------------------------------------------------------------------
-- AUTO-RANDOM FACTIONS
-------------------------------------------------------------------------------

function SpeakinSpell:InsertRandomFaction(name)
	if (name ~= nil) and (name ~= "") then
		tinsert(SpeakinSpellSavedData.RandomSubs["randomfaction"], name)
	end
end


function SpeakinSpell:AddRandomFactions()
	-- add the player's guild name (if any)
	-- guildName, guildRankName, guildRankIndex = GetGuildInfo(unit);
	local guildName = GetGuildInfo( UnitName("player") )
	self:InsertRandomFaction(guildName)
	
	-- add the player's arena teams (if any)
	-- teamName, teamSize, teamRating, weekPlayed, weekWins, seasonPlayed, seasonWins, playerPlayed, seasonPlayerPlayed, teamRank, playerRating = GetArenaTeam(id);
	local teamName2 = GetArenaTeam(1);
	local teamName3 = GetArenaTeam(2);
	local teamName5 = GetArenaTeam(5);

	self:InsertRandomFaction(teamName2)
	self:InsertRandomFaction(teamName3)
	self:InsertRandomFaction(teamName5)
end

